/*
   FNT - Daemon for handle xmpp-connections
   Main file

   Copyright (C) 2009 Free Software Foundation, Inc.

   Written by:
   Slava Zanko <slavazanko@gmail.com>


   This file is part of the FIDO New Technology.

   FIDO New Technology is free software: you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as published 
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#ifdef HAVE_SIGNAL_H
#	include <signal.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <gnutls/gnutls.h>
#include <gnutls/x509.h>


#include "../include/global.h"
#include "../include/types.h"
#include "../include/error.h"
#include "../include/sql.h"
#include "../include/main.h"
#include "../include/ssltls.h"
#include "../include/mainloop.h"
#include "../include/stream.h"


#include "settings.h"


#include "main.h"
#include "../include/sys-signal.h"

/*** global variables **************************************************/
fnt_mainloop_t *mainloop;
fnt_stream_t *server_stream;

fnt_sql_connect_t *data_connect = NULL;
fnt_sql_connect_t *config_connect = NULL;

/*** file scope macro definitions **************************************/

/*** file scope type declarations **************************************/

int xmppd_stream_callback_function (void *stream, int event, void *data);

/*** file scope variables **********************************************/

static const char *config_file = NULL;
static bool param_global = false;

static bool need_shutdown = false;
static bool need_reload_config = false;

xmppd_t *xmppd_app;

/*** file scope functions **********************************************/

static void
print_help (void)
{

    fprintf (stdout,
	     _(PACKAGE "-xmppd - daemon for handle xmpp-connections (%s)"),
	     VERSION);

    fprintf (stdout, "\n\nCopyright (C) 2009 Free Software Foundation, Inc.\n\
Written by: Slava Zanko <slavazanko@gmail.com>\n\
This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you\n\
are welcome to redistribute it under certain conditions.\n\n");
    fprintf (stdout, _("Usage:\n"));

    fprintf (stdout, PACKAGE "-xmppd\n\
" PACKAGE "-xmppd -h\n\
" PACKAGE "-xmppd -c <config>\n\
" PACKAGE "-xmppd -g\n\
");

    fprintf (stdout, _("Options are:\n\
   -c      Initial config file to use. Defaults:\n\
                   $FNT_INIT_FILE\n\
                   $HOME/.config/fnt.conf\n\
                   $HOME/.fnt.conf\n\
                   %s/fnt.conf.\n\
   -g              Use only global init file.\n\
   -h              Show this help.\n\n"), SYSCONF_DIR);

    exit (1);
}

static void
get_cmd_parameters (int *argc, char ***argv)
{
    int optchar;

    while ((optchar = getopt (*argc, *argv, "gdc:Li:h?")) >= 0)
    {
	switch (optchar)
	{
	case 'c':
	    config_file = optarg;
	    break;
	case 'g':
	    param_global = true;
	    break;
	case 'h':
	case '?':
	    print_help ();
	    break;
	}
    }
}

static void
check_cmd_parameters (char **argv)
{

}

static void
system_signal_shutdown (int signal)
{
    fnt_log (LOG_info, _("Exiting from main event loop"));
    fnt_mainloop_quit (mainloop);
}

static void
system_signal_SIGHUP (int signal)
{
    /* reread config, reinit application */
    need_reload_config = true;
}

static void
system_signals_setup (void)
{

    fnt_sys_signal (SIGINT, system_signal_shutdown);
    fnt_sys_signal (SIGTERM, system_signal_shutdown);
#ifdef SIGHUP
    fnt_sys_signal (SIGHUP, system_signal_SIGHUP);
#endif
#ifdef SIGPIPE
    fnt_sys_signal (SIGPIPE, SIG_IGN);
#endif

}

void
init_db_connects (void)
{
    char *error_str;

    if (!fnt_init ())
    {
	fprintf (stderr, _("ERROR: Unable initialize FNT library\n"));
	exit (2);
    }

    config_connect =
	fnt_getConfigConnectionFromIniFile (config_file, param_global);

    data_connect =
	fnt_getDataConnectionFromIniFile (config_file, param_global);


    if (config_connect == NULL || data_connect == NULL)
    {
	error_str = fnt_error_getStackAsString ();
	fprintf (stderr, "%s", error_str);
	g_free (error_str);
	if (config_connect)
	    fnt_sql_disconnect (&config_connect);
	if (data_connect)
	    fnt_sql_disconnect (&data_connect);
	fnt_deinit ();
	g_free (xmppd_app);
	exit (2);
    }

}

bool
init_log_connect (void)
{
    char *error_str;
    if (!fnt_log_init
	(settings->log_type, settings->log_level, PACKAGE "-xmppd",
	 (const char *) settings->facility_or_fname))
    {
	error_str = fnt_error_getStackAsString ();
	fprintf (stderr,
		 _
		 ("ERROR: Unable to init logging system. Logging disabled. \n%s"),
		 error_str);
	g_free (error_str);
    }
    fnt_log (LOG_notice, _("Starting"));

}

void
xmppd_deinit (void)
{
    pidfile_remove ();
    cleanup_settings ();
    fnt_log_deinit ();
    if (config_connect)
	fnt_sql_disconnect (&config_connect);
    if (data_connect)
	fnt_sql_disconnect (&data_connect);

    g_free (xmppd_app);
    fnt_deinit ();

}

char *
xmppd_getKeyPath (bool is_private)
{
    char *key, *tmp;
    if (is_private)
    {
	if (*settings->ssl_private_key != '\0')
	    return g_strdup (settings->ssl_private_key);

	tmp = g_strdup_printf ("%s.key", settings->fidonet_node_address);
	key = g_build_filename (settings->ssl_directory, "private", tmp,
				(char *) NULL);
    }
    else
    {
	if (*settings->ssl_public_key != '\0')
	    return g_strdup (settings->ssl_public_key);

	tmp = (char *) g_strdup_printf ("%s.crt", settings->fidonet_node_address);
	key = g_build_filename (settings->ssl_directory, tmp, (char *) NULL);
    }
    g_free (tmp);
    return key;
}

bool
xmppd_tls_init (void)
{
    bool ret;
    char *private_key;
    char *public_key;

    private_key = xmppd_getKeyPath (true);
    public_key = xmppd_getKeyPath (false);
    ret = fnt_mainloop_tls_init (mainloop,
				 settings->tls_auth_type,
				 (const char *) private_key,
				 (const char *) public_key);

    g_free (private_key);
    g_free (public_key);
    return ret;
}

bool
xmppd_init_server_stream()
{
    char *init_stream_data;
    server_stream = fnt_stream_init ("socket");
    if (server_stream == NULL)
    {
	fnt_log (LOG_error, _("Unable to create socket!"));
	return false;
    }
    init_stream_data = g_strdup_printf("server:%s:%d",settings->listen_ip,settings->listen_port);
    if (!fnt_stream_open (server_stream, init_stream_data))
    {
	fnt_log (LOG_error, _("Unable to listen socket!"));
	g_free(init_stream_data);
	fnt_stream_deinit (&server_stream);
	return false;
    }
    g_free(init_stream_data);
    fnt_stream_assignToMainloop (server_stream, mainloop);
    return true;
}

/*** public functions **************************************************/

int
main (int argc, char **argv)
{
    setlocale (LC_ALL, "");
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);

    get_cmd_parameters (&argc, &argv);
    check_cmd_parameters (argv);

    xmppd_app = g_try_malloc (sizeof (xmppd_t));
    if (xmppd_app == NULL)
    {
	fprintf (stderr, _("Unable allocate memory for struct '%s'\n"),
		 "xmppd_t");
	exit (2);
    }

    init_db_connects ();
    system_signals_setup ();
    read_settings ();
    init_log_connect ();
    pidfile_touch ();
    // ----------------------------------------------------


    mainloop = fnt_mainloop_init ();
    if (mainloop == NULL)
    {
	fnt_log (LOG_fatal, _("Error occured while starting program"));
	
    }


    if (! xmppd_tls_init ()){
	fnt_log (LOG_fatal, _("Error occured while starting program"));
	xmppd_deinit ();
	exit(1);
    }

    if (!fnt_channel_init(mainloop)){
	fnt_log (LOG_fatal, _("Fail to init channel subsystem!"));
	xmppd_deinit ();
	exit(1);
    }

    if (!xmppd_init_server_stream())
    {
	fnt_log (LOG_fatal, _("Fail to init server socket!"));
	xmppd_deinit ();
	exit(1);
    }

    fnt_log (LOG_info, _("Running main event loop"));

    fnt_mainloop_run (mainloop);



    fnt_stream_close (server_stream);
    fnt_stream_deinit (&server_stream);

    fnt_mainloop_deinit (&mainloop);

    fnt_log (LOG_notice, _("Shutting down"));
    xmppd_deinit ();
    return 0;
}
